@@ -1,14 +1,16 @@
 from __future__ import print_function
+from collections import defaultdict
+import itertools
 from prompt_toolkit.completion import Completer, Completion
-import sqlparse
+from .packages.sqlcompletion import suggest_type
 
 class PGCompleter(Completer):
     keywords = ['ACCESS', 'ADD', 'ALL', 'ALTER TABLE', 'AND', 'ANY', 'AS',
-            'ASC', 'AUDIT', 'BETWEEN', 'BY', 'CHAR', 'CHECK', 'CLUSTER',
+            'ASC', 'AUDIT', 'BETWEEN', 'BY', 'CASE', 'CHAR', 'CHECK', 'CLUSTER',
             'COLUMN', 'COMMENT', 'COMPRESS', 'CONNECT', 'CREATE', 'CURRENT',
             'DATE', 'DECIMAL', 'DEFAULT', 'DELETE FROM', 'DESC', 'DESCRIBE',
             'DISTINCT', 'DROP', 'ELSE', 'EXCLUSIVE', 'EXISTS', 'FILE', 'FLOAT',
-            'FOR', 'FROM', 'GRANT', 'GROUP', 'HAVING', 'IDENTIFIED',
+            'FOR', 'FROM', 'FULL', 'GRANT', 'GROUP BY', 'HAVING', 'IDENTIFIED',
             'IMMEDIATE', 'IN', 'INCREMENT', 'INDEX', 'INITIAL', 'INSERT INTO',
             'INTEGER', 'INTERSECT', 'INTO', 'IS', 'LEVEL', 'LIKE', 'LOCK',
             'LONG', 'MAXEXTENTS', 'MINUS', 'MLSLABEL', 'MODE', 'MODIFY',
@@ -19,13 +21,19 @@ class PGCompleter(Completer):
             'SET', 'SHARE', 'SIZE', 'SMALLINT', 'START', 'SUCCESSFUL',
             'SYNONYM', 'SYSDATE', 'TABLE', 'THEN', 'TO', 'TRIGGER', 'UID',
             'UNION', 'UNIQUE', 'UPDATE', 'USE', 'USER', 'VALIDATE', 'VALUES',
-            'VARCHAR', 'VARCHAR2', 'VIEW', 'WHENEVER', 'WHERE', 'WITH', ]
+            'VARCHAR', 'VARCHAR2', 'VIEW', 'WHEN', 'WHENEVER', 'WHERE', 'WITH', ]
+
+    functions = ['AVG', 'COUNT', 'DISTINCT', 'FIRST', 'FORMAT', 'LAST', 'LCASE',
+                 'LEN', 'MAX', 'MIN', 'MID', 'NOW', 'ROUND', 'SUM', 'TOP',
+                 'UCASE']
 
     special_commands = []
 
-    database_names = []
-    table_names = []
-    column_names = ['*']
+    databases = []
+    tables = []
+    # This will create a defaultdict which is initialized with a list that has
+    # a '*' by default.
+    columns = defaultdict(lambda: ['*'])
     all_completions = set(keywords)
 
     def __init__(self, smart_completion=True):
@@ -37,24 +45,24 @@ class PGCompleter(Completer):
         # be at the beginning of a line.
         self.special_commands.extend(special_commands)
 
-    def extend_database_names(self, database_names):
-        self.database_names.extend(database_names)
+    def extend_database_names(self, databases):
+        self.databases.extend(databases)
 
     def extend_keywords(self, additional_keywords):
         self.keywords.extend(additional_keywords)
         self.all_completions.update(additional_keywords)
 
-    def extend_table_names(self, table_names):
-        self.table_names.extend(table_names)
-        self.all_completions.update(table_names)
+    def extend_table_names(self, tables):
+        self.tables.extend(tables)
+        self.all_completions.update(tables)
 
-    def extend_column_names(self, column_names):
-        self.column_names.extend(column_names)
-        self.all_completions.update(column_names)
+    def extend_column_names(self, table, columns):
+        self.columns[table].extend(columns)
+        self.all_completions.update(columns)
 
     def reset_completions(self):
-        self.table_names = []
-        self.column_names = ['*']
+        self.tables = []
+        self.columns = defaultdict(itertools.repeat(['*']).next)
         self.all_completions = set(self.keywords)
 
     @staticmethod
@@ -72,32 +80,26 @@ class PGCompleter(Completer):
         if not self.smart_completion:
             return self.find_matches(word_before_cursor, self.all_completions)
 
-        # If we've partially typed a word then word_before_cursor won't be an
-        # empty string. In that case we want to remove the partially typed
-        # string before sending it to the sqlparser. Otherwise the last token
-        # will always be the partially typed string which renders the smart
-        # completion useless because it will always return the list of keywords
-        # as completion.
-        if word_before_cursor:
-            parsed = sqlparse.parse(
-                    document.text_before_cursor[:-len(word_before_cursor)])
-        else:
-            parsed = sqlparse.parse(document.text_before_cursor)
+        category, scope = suggest_type(document.text,
+                document.text_before_cursor)
 
-        last_token = ''
-        if parsed:
-            last_token = parsed[0].token_prev(len(parsed[0].tokens))
-            last_token = last_token.value if last_token else ''
-
-        if last_token.lower() in ('select', 'where', 'having', 'set',
-                'order by', 'group by'):
-            return self.find_matches(word_before_cursor, self.column_names)
-        elif last_token.lower() in ('from', 'update', 'into', 'describe'):
-            return self.find_matches(word_before_cursor, self.table_names)
-        elif last_token.lower() in ('d',):  # This for the \d special command.
-            return self.find_matches(word_before_cursor, self.table_names)
-        elif last_token.lower() in ('c', 'use'):  # This for the \c special command.
-            return self.find_matches(word_before_cursor, self.database_names)
-        else:
-            return self.find_matches(word_before_cursor,
-                    self.keywords + self.special_commands)
+        if category == 'columns':
+            scoped_cols = []
+            for table in scope:
+                scoped_cols.extend(self.columns[table])
+            print ('scope:', scope)
+            print ('scoped_cols:', scoped_cols)
+            return self.find_matches(word_before_cursor, scoped_cols)
+        elif category == 'columns-and-functions':
+            scoped_cols = []
+            for table in scope:
+                scoped_cols.extend(self.columns[table])
+            return self.find_matches(word_before_cursor, scoped_cols +
+                    self.functions)
+        elif category == 'tables':
+            return self.find_matches(word_before_cursor, self.tables)
+        elif category == 'databases':
+            return self.find_matches(word_before_cursor, self.databases)
+        elif category == 'keywords':
+            return self.find_matches(word_before_cursor, self.keywords +
+                    self.special_commands)
